import json
from pathlib import Path

import numpy as np
import torch
from loguru import logger

from .. import natural_sort
from ..data.seasonality import day_of_year_cyclic_feats
from ..train.model import CloudNet


def load_model_files(
    model_path: str,
    idx=-1,
    load_last=False,
    model_config_only=False,
    classes=list(range(4)),
):
    """
    Load model from model path and return model, model_class_mapping, config
    """

    with open(f"{model_path}/config.json", "r") as file:
        model_config = json.load(file)
    input_bands = [band[4:] for band in model_config["bands_config"]["s2_bands"]]
    model_config["bands_config"]["s2_bands"] = input_bands

    if model_config["data_config"]["classify_snow"]:
        model_class_mapping = model_config["labels_config"][
            "cloudsen12_mergedclouds_extrasnow"
        ]
    else:
        model_class_mapping = model_config["labels_config"]["cloudsen12"]

    if not model_path.endswith("/"):
        model_path += "/"

    model_checkpoint_path = model_path + "checkpoints/"
    logger.debug(f"Model checkpoint path: {model_checkpoint_path}")

    assert Path(model_checkpoint_path).exists(), (
        f"Model checkpoint path {model_checkpoint_path} does not exist"
    )

    if load_last:
        model_checkpoint = Path(model_checkpoint_path) / "last.ckpt"
    else:
        available_checkpoints = natural_sort(
            Path(model_checkpoint_path).glob("ep*.ckpt")
        )
        logger.debug(f"Available checkpoints: {available_checkpoints}")
        model_checkpoint = available_checkpoints[idx]
        logger.debug(f"Loading model from {model_checkpoint}")

    model_state_dict = torch.load(
        model_checkpoint,
        map_location="cpu",
        weights_only=True,  # NOTE define weights for better understanding
    )["state_dict"]
    logger.info(f"Model state dict loaded from {model_checkpoint}")

    if model_config_only:
        logger.info("Returning model configuration only")
        return model_config

    logger.info("Initializing CloudNet model")
    model = CloudNet(
        bands=model_config["bands_config"]["s2_bands"],
        classes=classes,
        arch=model_config["dl_model_config"]["arch"],
        backbone=model_config["dl_model_config"]["backbone"],
        activation=model_config["dl_model_config"]["activation"],
        loss=model_config["dl_model_config"]["loss"],
        learning_rate=model_config["dl_model_config"]["learning_rate"],
        class_weights_list=model_config["dl_model_config"]["class_weights_list"],
        sample_weights=model_config["dl_model_config"]["sample_weights"],
        label_smoothing_factor=model_config["dl_model_config"][
            "label_smoothing_factor"
        ],
        dropout_rate=model_config["dl_model_config"]["dropout_rate"],
        config=model_config,
        architecture_version=model_config["dl_model_config"]["architecture_version"],
        mlp_hidden_layers=model_config["dl_model_config"]["mlp_hidden_layers"],
        head_filters_settings=model_config["dl_model_config"]["head_filters_settings"],
        bands_head=model_config["dl_model_config"]["bands_head"],
    )
    model.load_state_dict(model_state_dict, strict=False)
    logger.info("Model state dict loaded successfully")

    model.eval()
    logger.info("Model set to evaluation mode")
    return model, model_class_mapping, model_config


def get_feats_head(shape_X, epsg, bounds, date, model_config):
    """
    Generate the feature head for the model

    Args:
    - shape_X: shape of the input tensor
    - epsg: EPSG code of the input data
    - bounds: bounds of the input data
    - date: date of the input data
    - model_config: model configuration

    Returns:
    - feats_head: feature head for the model
    """

    resolution = model_config["data_config"]["resolution"]
    try:
        year = int(str(date)[:4])
    except Exception as e:
        logger.error(f"Error in extracting year from date: {date}, {e}")

    feats_head = []
    if "meteo" in model_config["dl_model_config"]["bands_head"]:
        from evotrain.meteo import load_meteo_embeddings

        logger.info("Loading meteo embeddings")
        meteo_emb = load_meteo_embeddings(
            bounds, epsg, year, bounds_buffer=3000, order=2, resolution=resolution
        )
        logger.debug(f"Meteo embeddings loaded with shape: {meteo_emb.shape}")

        meteo_emb /= 250
        logger.debug("Meteo embeddings scaled to 0-1 range")
        feats_head.append(meteo_emb)

    if "latlon" in model_config["dl_model_config"]["bands_head"]:
        from evotrain.models.lcm10.v12.inference import (
            lat_lon_to_unit_sphere,
            load_latlon,
        )

        logger.info("Loading latlon and converting to unit sphere (x,y,z)")
        lat, lon = load_latlon(bounds, epsg, resolution=resolution)
        xx, yy, zz = lat_lon_to_unit_sphere(lat, lon)
        logger.debug("Latlon converted to unit sphere")

        xx = np.full((1, shape_X[1], shape_X[2]), xx)
        yy = np.full((1, shape_X[1], shape_X[2]), yy)
        zz = np.full((1, shape_X[1], shape_X[2]), zz)
        logger.debug(
            f"Unit sphere arrays created with shapes: {xx.shape}, {yy.shape}, {zz.shape}"
        )
        feats_head.append(xx)
        feats_head.append(yy)
        feats_head.append(zz)

    if "seasonal" in model_config["dl_model_config"]["bands_head"]:
        logger.info("Generating seasonal features")
        sin_cos = day_of_year_cyclic_feats(
            str(date), doy_jitter=15, height=shape_X[1], width=shape_X[2]
        )
        # Let's scale the sin_cos features to the range [0, 1]
        sin_cos = (sin_cos + 1) / 2
        logger.debug(f"Seasonal features generated with shape: {sin_cos.shape}")
        feats_head.append(sin_cos)

    feats_head = np.concatenate(feats_head, axis=0).astype(np.float32)
    logger.info(f"Feature head generated with shape: {feats_head.shape}")

    return feats_head
